import argparse
import asyncio
from collections import Counter, defaultdict

import copy
import discord
from discord.ext import commands

import discordbot.embeds
from ..bot_utils import checks
from ..bot_utils import config
from ..colors import Colors


class Arguments(argparse.ArgumentParser):
    def error(self, message):
        raise RuntimeError(message)


class BotAdmin:
    """Bot administration commands."""

    def __init__(self, bot):
        self.bot = bot
        self.config = config.Config('botadmin.json', loop=bot.loop, directory="data")

        # guild_id: set(user_id)
        self._recently_kicked = defaultdict(set)

    def bot_user(self, message):
        return message.server.me if message.channel.is_private else self.bot.user

    def is_plonked(self, server, member):
        db = self.config.get('plonks', {}).get(server.id, [])
        bypass_ignore = member.server_permissions.manage_server
        if not bypass_ignore and member.id in db:
            return True
        return False

    def __check(self, ctx):
        msg = ctx.message

        if checks.is_owner_check(msg):
            return True

        # user is bot banned
        if msg.server:
            if self.is_plonked(msg.server, msg.author):
                return False

        # check if the channel is ignored
        # but first, resolve their permissions

        perms = msg.channel.permissions_for(msg.author)
        bypass_ignore = perms.administrator

        # now we can finally realise if we can actually bypass the ignore.

        if not bypass_ignore and msg.channel.id in self.config.get('ignored', []):
            return False

        if checks.is_owner_check(msg):
            return True

        try:
            entry = self.config.get('commands', {}).get(msg.server.id, [])
        except (KeyError, AttributeError):
            return True
        else:
            name = ctx.command.qualified_name.split(' ')[0]
            return name not in entry

    async def on_message(self, message):
        if message.author.bot:
            return

        if message.channel.is_private and not message.content.startswith(self.bot.command_prefix):
            owner_id = self.bot.config.get("meta", {}).get("owner", "249746236008169473")
            to_send = self.bot.config.get("meta", {}).get("send_dms", True)
            image = ""
            embed = {}
            if message.embeds:
                embed = message.embeds[0]
            if message.attachments:
                image = message.attachments[0]['url']
            if message.author.id != owner_id and to_send:
                owner = await self.bot.get_user_info(owner_id)
                if message.content:
                    await self.bot.responses.basic(destination=owner, message=message.content, author=str(message.author), author_img=message.author.avatar_url, color=Colors.get_default(self.bot), image=image)
                if embed:
                    embed.pop("author", "")
                    url = embed.pop("thumbnail", {}).get("url", "")
                    e = discord.Embed.from_data(embed)
                    e.color = Colors.get_default(self.bot)
                    e.set_image(url=url)
                    e.set_author(name="Embed from " + str(message.author), icon_url=message.author.avatar_url)
                    await self.bot.send_message(destination=owner, embed=e)

    @commands.command(name='dms', hidden=True)
    @checks.is_owner()
    async def _senddms(self):
        """Toggles sending DMs to owner."""
        data = self.bot.config.get("meta", {})
        tosend = data.get('send_dms', True)
        data['send_dms'] = not tosend
        await self.bot.config.put('meta', data)
        await self.bot.responses.toggle(message="Forwarding of DMs to owner has been {status}.", success=data['send_dms'])

    @commands.command(name='quit', hidden=True)
    @checks.is_owner()
    async def _quit(self):
        """Quits the bot."""
        await self.bot.responses.failure(message="Bot shutting down")
        await self.bot.logout()

    @commands.command(name='setname', hidden=True)
    @checks.is_owner()
    async def _setname(self, *, username: str):
        """Changes the bots username."""
        await self.bot.edit_profile(username=username)

    @commands.command(name='setcolor', hidden=True)
    @checks.is_owner()
    async def _setcolor(self, *, color : discord.Colour):
        """Sets the default color of embeds."""
        data = self.bot.config.get("meta", {})
        data['default_color'] = str(color)
        await self.bot.config.put('meta', data)
        await self.bot.responses.basic(message="The default color has been updated.")

    @commands.command(name='runtest', hidden=True)
    @checks.is_owner()
    async def _runtest(self):
        """Sets the default color of embeds."""
        sections = [{"name": "Section 1", "value": "Value 1"}, {"name": "Section 2"}, {"name": "Section 2.5", "value": "Value 2.5"},
                    {"name": "Section 3", "value": "Value 3", "inline": False}]
        e = discordbot.embeds.build_embed(title="IDK", description="foo bar", sections=sections)
        await self.bot.say(embed=e)

    @discordbot.command(name="do", pass_context=True, hidden=True)
    @discordbot.checks.is_owner()
    async def _do(self, ctx, times: int, *, command):
        """Repeats a command a specified number of times."""
        msg = copy.copy(ctx.message)
        msg.content = command
        for i in range(times):
            await self.bot.process_commands(msg)

    @commands.command(pass_context=True, no_pm=True)
    @checks.admin_or_permissions(manage_server=True)
    async def disable(self, ctx, *, command: str):
        """Disables a command for this server.

        You must have Manage Server permissions or the
        Bot Admin role to use this command.
        """
        command = command.lower()

        if command in ('enable', 'disable'):
            return await self.bot.responses.failure(message='Cannot disable that command.')

        if command not in self.bot.commands:
            return await self.bot.responses.failure(message='Command "{}" was not found.'.format(command))

        guild_id = ctx.message.server.id
        cmds = self.config.get('commands', {})
        entries = cmds.get(guild_id, [])
        entries.append(command)
        cmds[guild_id] = entries
        await self.config.put('commands', cmds)
        await self.bot.responses.success(message='"%s" command disabled in this server.' % command)

    @commands.command(pass_context=True, no_pm=True)
    @checks.admin_or_permissions(manage_server=True)
    async def enable(self, ctx, *, command: str):
        """Enables a command for this server.

        You must have Manage Server permissions or the
        Bot Admin role to use this command.
        """
        command = command.lower()
        guild_id = ctx.message.server.id
        cmds = self.config.get('commands', {})
        entries = cmds.get(guild_id, [])

        try:
            entries.remove(command)
        except KeyError:
            await self.bot.responses.failure(message='The command does not exist or is not disabled.')
        else:
            cmds[guild_id] = entries
            await self.config.put('commands', cmds)
            await self.bot.responses.success(message='"%s" command enabled in this server.' % command)

    @commands.group(pass_context=True, no_pm=True)
    @checks.admin_or_permissions(manage_channels=True)
    async def ignore(self, ctx):
        """Handles the bot's ignore lists.

        To use these commands, you must have the Bot Admin role or have
        Manage Channels permissions. These commands are not allowed to be used
        in a private message context.

        Users with Manage Roles or Bot Admin role can still invoke the bot
        in ignored channels.
        """
        if ctx.invoked_subcommand is None:
            await self.bot.say('Invalid subcommand passed: {0.subcommand_passed}'.format(ctx))

    @ignore.command(name='list', pass_context=True)
    async def ignore_list(self, ctx):
        """Tells you what channels are currently ignored in this server."""

        ignored = self.config.get('ignored', [])
        channel_ids = set(c.id for c in ctx.message.server.channels)
        result = []
        for channel in ignored:
            if channel in channel_ids:
                result.append('<#{}>'.format(channel))

        if result:
            await self.bot.responses.basic(title="Ignored Channels:", message='\n\n{}'.format(', '.join(result)))
        else:
            await self.bot.responses.failure(message='I am not ignoring any channels here.')

    @ignore.command(name='channel', pass_context=True)
    async def channel_cmd(self, ctx, *, channel : discord.Channel = None):
        """Ignores a specific channel from being processed.

        If no channel is specified, the current channel is ignored.
        If a channel is ignored then the bot does not process commands in that
        channel until it is unignored.
        """

        if channel is None:
            channel = ctx.message.channel

        ignored = self.config.get('ignored', [])
        if channel.id in ignored:
            await self.bot.responses.failure(message='That channel is already ignored.')
            return

        ignored.append(channel.id)
        await self.config.put('ignored', ignored)
        await self.bot.responses.success(message='Channel <#{}> will be ignored.'.format(channel.id))

    @ignore.command(name='all', pass_context=True)
    @checks.admin_or_permissions(manage_server=True)
    async def _all(self, ctx):
        """Ignores every channel in the server from being processed.

        This works by adding every channel that the server currently has into
        the ignore list. If more channels are added then they will have to be
        ignored by using the ignore command.

        To use this command you must have Manage Server permissions along with
        Manage Channels permissions. You could also have the Bot Admin role.
        """

        ignored = self.config.get('ignored', [])
        channels = ctx.message.server.channels
        ignored.extend(c.id for c in channels if c.type == discord.ChannelType.text)
        await self.config.put('ignored', list(set(ignored))) # make unique
        await self.bot.responses.success(message='All channels ignored.')

    @commands.group(pass_context=True, no_pm=True, invoke_without_command=True)
    @checks.admin_or_permissions(manage_channels=True)
    async def unignore(self, ctx, *channels: discord.Channel):
        """Unignores channels from being processed.

        If no channels are specified, it unignores the current channel.

        To use this command you must have the Manage Channels permission or have the
        Bot Admin role.
        """

        if len(channels) == 0:
            channels = (ctx.message.channel,)

        # a set is the proper data type for the ignore list
        # however, JSON only supports arrays and objects not sets.
        ignored = self.config.get('ignored', [])
        result = []
        for channel in channels:
            try:
                ignored.remove(channel.id)
            except ValueError:
                pass
            else:
                result.append('<#{}>'.format(channel.id))

        await self.config.put('ignored', ignored)
        await self.bot.responses.success(message='Channel(s) {} will no longer be ignored.'.format(', '.join(result)))

    @unignore.command(name='all', pass_context=True, no_pm=True)
    @checks.admin_or_permissions(manage_channels=True)
    async def unignore_all(self, ctx):
        """Unignores all channels in this server from being processed.

        To use this command you must have the Manage Channels permission or have the
        Bot Admin role.
        """
        channels = [c for c in ctx.message.server.channels if c.type is discord.ChannelType.text]
        await ctx.invoke(self.unignore, *channels)

    @commands.command(pass_context=True, no_pm=True)
    @checks.mod_or_permissions(manage_messages=True)
    async def cleanup(self, ctx, search : int = 100):
        """Cleans up the bot's messages from the channel.

        If a search number is specified, it searches that many messages to delete.
        If the bot has Manage Messages permissions, then it will try to delete
        messages that look like they invoked the bot as well.

        After the cleanup is completed, the bot will send you a message with
        which people got their messages deleted and their count. This is useful
        to see which users are spammers.

        To use this command you must have Manage Messages permission or have the
        Bot Mod role.
        """

        spammers = Counter()
        channel = ctx.message.channel
        prefixes = self.bot.command_prefix
        if callable(prefixes):
            prefixes = prefixes(self.bot, ctx.message)

        def is_possible_command_invoke(entry):
            valid_call = any(entry.content.startswith(prefix) for prefix in prefixes)
            return valid_call and not entry.content[1:2].isspace()

        can_delete = channel.permissions_for(channel.server.me).manage_messages

        if not can_delete:
            api_calls = 0
            async for entry in self.bot.logs_from(channel, limit=search, before=ctx.message):
                if api_calls and api_calls % 5 == 0:
                    await asyncio.sleep(1.1)

                if entry.author == self.bot.user:
                    await self.bot.delete_message(entry)
                    spammers['Bot'] += 1
                    api_calls += 1

                if is_possible_command_invoke(entry):
                    try:
                        await self.bot.delete_message(entry)
                    except discord.Forbidden:
                        continue
                    else:
                        spammers[entry.author.display_name] += 1
                        api_calls += 1
        else:
            predicate = lambda m: m.author == self.bot.user or is_possible_command_invoke(m)
            deleted = await self.bot.purge_from(channel, limit=search, before=ctx.message, check=predicate)
            spammers = Counter(m.author.display_name for m in deleted)

        deleted = sum(spammers.values())
        messages = ['%s %s removed.' % (deleted, 'message was' if deleted == 1 else 'messages were')]
        if deleted:
            messages.append('')
            spammers = sorted(spammers.items(), key=lambda t: t[1], reverse=True)
            messages.extend(map(lambda t: '**{0[0]}**: {0[1]}'.format(t), spammers))

        msg = await self.bot.responses.basic(title="Removed Messages:", message='\n'.join(messages))
        await asyncio.sleep(10)
        await self.bot.delete_message(msg)

    @commands.command(no_pm=True, pass_context=True)
    @checks.admin_or_permissions(manage_server=True)
    async def plonk(self, ctx, *, member: discord.Member):
        """Bans a user from using the bot.

        This bans a person from using the bot in the current server.
        There is no concept of a global ban. This ban can be bypassed
        by having the Manage Server permission.

        To use this command you must have the Manage Server permission
        or have a Bot Admin role.
        """

        plonks = self.config.get('plonks', {})
        guild_id = ctx.message.server.id
        db = plonks.get(guild_id, [])

        if member.id in db:
            await self.bot.responses.failure(message='That user is already bot banned in this server.')
            return

        db.append(member.id)
        plonks[guild_id] = db
        await self.config.put('plonks', plonks)
        await self.bot.responses.success(message='%s has been banned from using the bot in this server.' % member)

    @commands.command(no_pm=True, pass_context=True)
    @checks.admin_or_permissions(manage_server=True)
    async def plonks(self, ctx):
        """Shows members banned from the bot."""
        plonks = self.config.get('plonks', {})
        guild = ctx.message.server
        db = plonks.get(guild.id, [])
        members = '\n'.join(map(str, filter(None, map(guild.get_member, db))))
        if members:
            await self.bot.responses.basic(title="Plonked Users:", message=members)
        else:
            await self.bot.responses.failure(message='No members are banned in this server.')

    @commands.command(no_pm=True, pass_context=True)
    @checks.admin_or_permissions(manage_server=True)
    async def unplonk(self, ctx, *, member: discord.Member):
        """Unbans a user from using the bot.

        To use this command you must have the Manage Server permission
        or have a Bot Admin role.
        """

        plonks = self.config.get('plonks', {})
        guild_id = ctx.message.server.id
        db = plonks.get(guild_id, [])

        try:
            db.remove(member.id)
        except ValueError:
            await self.bot.responses.failure(message='%s is not banned from using the bot in this server.' % member)
        else:
            plonks[guild_id] = db
            await self.config.put('plonks', plonks)
            await self.bot.responses.success(message='%s has been unbanned from using the bot in this server.' % member)

def setup(bot):
    bot.add_cog(BotAdmin(bot))



import json
import os
import uuid
import asyncio

class Config:
    """The "database" object. Internally based on ``json``."""

    def __init__(self, name, **options):
        self.name = name
        self.object_hook = options.pop('object_hook', None)
        self.folder = options.pop('directory', "data")

        self.encoder = options.pop('encoder', None)
        self.loop = options.pop('loop', asyncio.get_event_loop())
        self.lock = asyncio.Lock()
        if self.folder:
            self.folder += "/"
            os.makedirs(self.folder, exist_ok=True)
        if options.pop('load_later', False):
            self.loop.create_task(self.load())
        else:
            self.load_from_file()

    def load_from_file(self):
        try:
            with open(self.folder+self.name, 'r') as f:
                self._db = json.load(f, object_hook=self.object_hook)
        except FileNotFoundError:
            self._db = {}

    async def load(self):
        with await self.lock:
            await self.loop.run_in_executor(None, self.load_from_file)

    def _dump(self):
        temp = '%s-%s.tmp' % (self.folder+self.name, uuid.uuid4())
        with open(temp, 'w', encoding='utf-8') as tmp:
            json.dump(self._db.copy(), tmp, ensure_ascii=True, cls=self.encoder, separators=(',', ':'))

        # automatically move the file
        os.replace(temp, self.folder+self.name)

    async def save(self):
        with await self.lock:
            await self.loop.run_in_executor(None, self._dump)

    def get(self, key, *args):
        """Retrieves a data entry."""
        return self._db.get(key, *args)

    async def put(self, key, value, *args):
        """Edits a data entry."""
        self._db[key] = value
        await self.save()

    async def remove(self, key):
        """Removes a data entry."""
        del self._db[key]
        await self.save()

    def __contains__(self, item):
        return item in self._db

    def __getitem__(self, item):
        return self._db[item]

    def __len__(self):
        return len(self._db)

    def all(self):
        return self._db



import copy, os
import datetime
import traceback
from collections import Counter
from collections import OrderedDict

import discord
import psutil
from discord.ext import commands

from ..bot_utils import config, checks
from ..bot_utils.paginator import Pages
from ..colors import Colors


class Meta:
    """Commands for utilities related to Discord or the Bot itself."""

    def __init__(self, bot):
        self.bot = bot
        self.process = psutil.Process()
        self.config = config.Config('stats.json', loop=bot.loop, directory="data")

    @commands.command(pass_context=True, aliases=['invite'])
    async def join(self, ctx):
        """Sends you the bot invite link."""
        perms = discord.Permissions.none()
        perms.read_messages = True
        perms.send_messages = True
        perms.manage_messages = True
        perms.embed_links = True
        perms.read_message_history = True
        perms.attach_files = True
        perms.add_reactions = True
        await self.bot.send_message(ctx.message.author, discord.utils.oauth_url(self.bot.client_id, perms))

    @commands.group(pass_context=True, no_pm=True, invoke_without_command=True)
    async def info(self, ctx, *, member : discord.Member = None):
        """Shows info about a member.

        This cannot be used in private messages. If you don't specify
        a member then the info returned will be yours.
        """
        channel = ctx.message.channel
        if member is None:
            member = ctx.message.author

        e = discord.Embed()
        roles = [role.name.replace('@', '@\u200b') for role in member.roles]
        shared = sum(1 for m in self.bot.get_all_members() if m.id == member.id)
        voice = member.voice_channel
        if voice is not None:
            other_people = len(voice.voice_members) - 1
            voice_fmt = '{} with {} others' if other_people else '{} by themselves'
            voice = voice_fmt.format(voice.name, other_people)
        else:
            voice = 'Not connected.'

        e.set_author(name=str(member), icon_url=member.avatar_url or member.default_avatar_url)
        e.set_footer(text='Member since').timestamp = member.joined_at
        e.add_field(name='ID', value=member.id)
        e.add_field(name='Servers', value='%s shared' % shared)
        e.add_field(name='Voice', value=voice)
        e.add_field(name='Created', value=member.created_at)
        e.add_field(name='Roles', value=', '.join(roles))
        e.colour = member.colour

        if member.avatar:
            e.set_image(url=member.avatar_url)

        await self.bot.say(embed=e)

    @info.command(name='server', pass_context=True, no_pm=True)
    async def server_info(self, ctx):
        server = ctx.message.server
        roles = [role.name.replace('@', '@\u200b') for role in server.roles]

        secret_member = copy.copy(server.me)
        secret_member.id = '0'
        secret_member.roles = [server.default_role]

        # figure out what channels are 'secret'
        secret_channels = 0
        secret_voice = 0
        text_channels = 0
        for channel in server.channels:
            perms = channel.permissions_for(secret_member)
            is_text = channel.type == discord.ChannelType.text
            text_channels += is_text
            if is_text and not perms.read_messages:
                secret_channels += 1
            elif not is_text and (not perms.connect or not perms.speak):
                secret_voice += 1

        voice_channels = len(server.channels) - text_channels
        member_by_status = Counter(str(m.status) for m in server.members)

        e = discord.Embed()
        e.title = 'Info for ' + server.name
        e.add_field(name='ID', value=server.id)
        e.add_field(name='Owner', value=server.owner)
        if server.icon:
            e.set_thumbnail(url=server.icon_url)

        if server.splash:
            e.set_image(url=server.splash_url)

        e.add_field(name='Partnered?', value='Yes' if len(server.features) >= 3 else 'No')

        fmt = 'Text %s (%s secret)\nVoice %s (%s locked)'
        e.add_field(name='Channels', value=fmt % (text_channels, secret_channels, voice_channels, secret_voice))

        fmt = 'Total: {0}\nOnline: {1[online]}' \
              ', Offline: {1[offline]}' \
              '\nDnD: {1[dnd]}' \
              ', Idle: {1[idle]}'

        e.add_field(name='Members', value=fmt.format(server.member_count, member_by_status))
        e.add_field(name='Roles', value=', '.join(roles) if len(roles) < 10 else '%s roles' % len(roles))
        e.set_footer(text='Created').timestamp = server.created_at
        await self.bot.say(embed=e)

    async def on_command(self, command, ctx):
        message = ctx.message
        if message.channel.is_private:
            id = ctx.message.author.id
            destination = 'Private Message'
        else:
            id = ctx.message.server.id
            destination = '#{0.channel.name} ({0.server.name})'.format(message)

        self.bot.logs['stats'].info('{0.timestamp}: {0.author.name} in {1}: {0.content}'.format(message, destination))

        data = self.config.get('data', {})
        server_data = data.get(id, {})
        server_data[ctx.command.qualified_name] = server_data.get(ctx.command.qualified_name, 0) + 1
        data[id] = server_data
        await self.config.put('data', data)

    @commands.command(pass_context=True, hidden=True)
    @checks.is_owner()
    async def commandstats(self, ctx, all_stats : bool=False):
        data = self.config.get('data', {})
        if all_stats:
            title = "Bot"
            commands = {}
            for d in data:
                server_cmds = data.get(d, {})
                for c in server_cmds.keys():
                    current = commands.get(c, 0)
                    commands[c] = current + server_cmds[c]
        else:
            title = "This Server"
            commands = data.get(ctx.message.server.id, {})
        commands = OrderedDict(sorted(commands.items(), key=lambda x: x[1], reverse=True))

        stats = []
        for c in commands.keys():
            stats.append(c + ": " + str(commands[c]))
        try:
            pager = Pages(self.bot, message=ctx.message, entries=stats)
            pager.embed.colour = Colors.get_default(self.bot)
            pager.embed.set_author(name="Command stats for " + title,
                                   icon_url=ctx.message.server.icon_url)
            await pager.paginate()

        except Exception as e:
            await self.bot.say(e)

    def get_bot_uptime(self, *, brief=False):
        now = datetime.datetime.utcnow()
        delta = now - self.bot.uptime
        hours, remainder = divmod(int(delta.total_seconds()), 3600)
        minutes, seconds = divmod(remainder, 60)
        days, hours = divmod(hours, 24)

        if not brief:
            if days:
                fmt = '{d} days, {h} hours, {m} minutes, and {s} seconds'
            else:
                fmt = '{h} hours, {m} minutes, and {s} seconds'
        else:
            fmt = '{h}h {m}m {s}s'
            if days:
                fmt = '{d}d ' + fmt

        return fmt.format(d=days, h=hours, m=minutes, s=seconds)

    @commands.command(aliases=['botstats'])
    async def about(self):
        """Tells you information about the bot itself.

        This is more information about the bot, not really
        sure when this comes into play tbh, thats what I am
        testing right now.
        """

        cmd = r'git show -s -3 --format="%s (%cr)"'
        if os.name == 'posix':
            cmd = cmd.format(r'\`%h\`')
        else:
            cmd = cmd.format(r'`%h`')

        embed = discord.Embed()
        embed.colour = Colors.get_default(self.bot)

        embed.add_field(name="About "+self.bot.user.name+":", value=self.bot.description, inline=False)
        
        if os.popen('git rev-list --all --count').close() is None and os.popen('git rev-parse').close() is None:
            revision = os.popen(cmd).read().strip()
            embed.add_field(name="Latest Changes:", value=revision, inline=False)

        try:
            owner = self._owner
        except AttributeError:
            owner = self._owner = await self.bot.get_user_info(self.bot.config.get('meta', {}).get('owner', "249746236008169473"))

        if self.bot.ownerInviteLink:
            embed.title = owner.name+'\'s Server Invite'
            embed.url = self.bot.ownerInviteLink

        embed.set_author(name="Created by "+str(owner), icon_url=owner.avatar_url)

        # statistics
        total_members = sum(len(s.members) for s in self.bot.servers)
        total_online  = sum(1 for m in self.bot.get_all_members() if m.status != discord.Status.offline)
        unique_members = set(self.bot.get_all_members())
        unique_online = sum(1 for m in unique_members if m.status != discord.Status.offline)
        channel_types = Counter(c.type for c in self.bot.get_all_channels())
        voice = channel_types[discord.ChannelType.voice]
        text = channel_types[discord.ChannelType.text]

        members = '%s total\n%s online\n%s unique\n%s unique online' % (total_members, total_online, len(unique_members), unique_online)
        embed.add_field(name='Members', value=members)
        embed.add_field(name='Channels', value='{} total\n{} text\n{} voice'.format(text + voice, text, voice))
        memory_usage = self.process.memory_full_info().uss / 1024**2
        cpu_usage = self.process.cpu_percent() / psutil.cpu_count()
        embed.add_field(name='Process', value='{:.2f} MiB\n{:.2f}% CPU'.format(memory_usage, cpu_usage))
        embed.set_footer(text='Made with discord.py & discordbot.py', icon_url='http://i.imgur.com/5BFecvA.png')
        embed.timestamp = self.bot.uptime

        data = self.config.get('data', {})
        summation = 0
        for server in data:
            summation += sum(data[server].values())

        embed.add_field(name='Servers', value=str(len(self.bot.servers)))
        embed.add_field(name='Commands Run', value=summation)
        embed.add_field(name='Uptime', value=self.get_bot_uptime(brief=True))

        await self.bot.say(embed=embed)

    async def send_server_stat(self, server_event, server):

        bots = sum(m.bot for m in server.members)
        total = server.member_count
        online = sum(m.status is discord.Status.online for m in server.members)

        if server.me:
            time = server.me.joined_at
        else:
            time = datetime.datetime.now()

        message = '{0} at {1}: {2.name} (ID: {2.id}), Owner: {2.owner} (ID: {2.owner.id}), Members: {3}, '\
                  'Bots: {4} ({5:.2%}), Online: {6} ({7:.2%})'
        message = message.format(server_event, time, server, total, bots, bots / total, online, online / total)
        self.bot.logs['info'].info(message)

    async def on_server_join(self, server):
        await self.send_server_stat("Joined Server", server)

    async def on_server_remove(self, server):
        await self.send_server_stat("Left Server", server)

    async def on_command_error(self, error, ctx):
        ignored = (commands.NoPrivateMessage, commands.DisabledCommand, commands.CheckFailure,
                   commands.CommandNotFound, commands.UserInputError, discord.HTTPException)
        error = getattr(error, 'original', error)

        if isinstance(error, ignored):
            return

        if ctx.message.server:
            fmt = 'Channel: {0} (ID: {0.id})\nGuild: {1} (ID: {1.id})'
        else:
            fmt = 'Channel: {0} (ID: {0.id})'

        exc = traceback.format_exception(type(error), error, error.__traceback__, chain=False)
        description = '```py\n%s\n```' % ''.join(exc)
        time = datetime.datetime.utcnow()

        name = ctx.command.qualified_name
        author = '{0} (ID: {0.id})'.format(ctx.message.author)
        location = fmt.format(ctx.message.channel, ctx.message.server)

        message = '{0} at {1}: Called by: {2} in {3}. More info: {4}'.format(name, time, author, location, description)

        self.bot.logs['discord'].critical(message)


def setup(bot):
    bot.add_cog(Meta(bot))



import asyncio
import discord

class CannotPaginate(Exception):
    pass

class Pages:
    """Implements a paginator that queries the user for the
    pagination interface.

    Pages are 1-index based, not 0-index based.

    If the user does not reply within 2 minutes then the pagination
    interface exits automatically.

    Parameters
    ------------
    bot
        The bot instance.
    message
        The message that initiated this session.
    entries
        A list of entries to paginate.
    per_page
        How many entries show up per page.

    Attributes
    -----------
    embed: discord.Embed
        The embed object that is being used to send pagination info.
        Feel free to modify this externally. Only the description,
        footer fields, and colour are internally modified.
    permissions: discord.Permissions
        Our permissions for the channel.
    """
    def __init__(self, bot, *, message, entries, per_page=10):
        self.bot = bot
        self.entries = entries
        self.message = message
        self.author = message.author
        self.per_page = per_page
        pages, left_over = divmod(len(self.entries), self.per_page)
        if left_over:
            pages += 1
        self.maximum_pages = pages
        self.embed = discord.Embed()
        self.paginating = len(entries) > per_page
        self.reaction_emojis = [
            ('\N{BLACK LEFT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}', self.first_page),
            ('\N{BLACK LEFT-POINTING TRIANGLE}', self.previous_page),
            ('\N{BLACK RIGHT-POINTING TRIANGLE}', self.next_page),
            ('\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR}', self.last_page),
            ('\N{INPUT SYMBOL FOR NUMBERS}', self.numbered_page ),
            ('\N{BLACK SQUARE FOR STOP}', self.stop_pages),
            ('\N{INFORMATION SOURCE}', self.show_help),
        ]

        server = self.message.server
        if server is not None:
            self.permissions = self.message.channel.permissions_for(server.me)
        else:
            self.permissions = self.message.channel.permissions_for(self.bot.user)

        if not self.permissions.embed_links:
            raise CannotPaginate('Bot does not have embed links permission.')

    def get_page(self, page):
        base = (page - 1) * self.per_page
        return self.entries[base:base + self.per_page]

    async def show_page(self, page, *, first=False):
        self.current_page = page
        entries = self.get_page(page)
        p = []
        for t in enumerate(entries, 1 + ((page - 1) * self.per_page)):
            p.append('%s. %s' % t)

        self.embed.set_footer(text='Page %s/%s (%s entries)' % (page, self.maximum_pages, len(self.entries)))

        if not self.paginating:
            self.embed.description = '\n'.join(p)
            return await self.bot.send_message(self.message.channel, embed=self.embed)

        if not first:
            self.embed.description = '\n'.join(p)
            await self.bot.edit_message(self.message, embed=self.embed)
            return

        # verify we can actually use the pagination session
        if not self.permissions.add_reactions:
            raise CannotPaginate('Bot does not have add reactions permission.')

        if not self.permissions.read_message_history:
            raise CannotPaginate('Bot does not have Read Message History permission.')

        p.append('')
        p.append('Confused? React with \N{INFORMATION SOURCE} for more info.')
        self.embed.description = '\n'.join(p)
        self.message = await self.bot.send_message(self.message.channel, embed=self.embed)
        for (reaction, _) in self.reaction_emojis:
            if self.maximum_pages == 2 and reaction in ('\u23ed', '\u23ee'):
                # no |<< or >>| buttons if we only have two pages
                # we can't forbid it if someone ends up using it but remove
                # it from the default set
                continue

            await self.bot.add_reaction(self.message, reaction)

    async def checked_show_page(self, page):
        if page != 0 and page <= self.maximum_pages:
            await self.show_page(page)

    async def first_page(self):
        """goes to the first page"""
        await self.show_page(1)

    async def last_page(self):
        """goes to the last page"""
        await self.show_page(self.maximum_pages)

    async def next_page(self):
        """goes to the next page"""
        await self.checked_show_page(self.current_page + 1)

    async def previous_page(self):
        """goes to the previous page"""
        await self.checked_show_page(self.current_page - 1)

    async def show_current_page(self):
        if self.paginating:
            await self.show_page(self.current_page)

    async def numbered_page(self):
        """lets you type a page number to go to"""
        to_delete = []
        to_delete.append(await self.bot.send_message(self.message.channel, 'What page do you want to go to?'))
        msg = await self.bot.wait_for_message(author=self.author, channel=self.message.channel,
                                              check=lambda m: m.content.isdigit(), timeout=30.0)
        if msg is not None:
            page = int(msg.content)
            to_delete.append(msg)
            if page != 0 and page <= self.maximum_pages:
                await self.show_page(page)
            else:
                to_delete.append(await self.bot.say('Invalid page given. (%s/%s)' % (page, self.maximum_pages)))
                await asyncio.sleep(5)
        else:
            to_delete.append(await self.bot.send_message(self.message.channel, 'Took too long.'))
            await asyncio.sleep(5)

        try:
            await self.bot.delete_messages(to_delete)
        except Exception:
            pass

    async def show_help(self):
        """shows this message"""
        e = discord.Embed()
        messages = ['Welcome to the interactive paginator!\n']
        messages.append('This interactively allows you to see pages of text by navigating with ' \
                        'reactions. They are as follows:\n')

        for (emoji, func) in self.reaction_emojis:
            messages.append('%s %s' % (emoji, func.__doc__))

        e.description = '\n'.join(messages)
        e.colour =  0x738bd7 # blurple
        e.set_footer(text='We were on page %s before this message.' % self.current_page)
        await self.bot.edit_message(self.message, embed=e)

        async def go_back_to_current_page():
            await asyncio.sleep(60.0)
            await self.show_current_page()

        self.bot.loop.create_task(go_back_to_current_page())

    async def stop_pages(self):
        """stops the interactive pagination session"""
        await self.bot.delete_message(self.message)
        self.paginating = False

    def react_check(self, reaction, user):
        if user is None or user.id != self.author.id:
            return False

        for (emoji, func) in self.reaction_emojis:
            if reaction.emoji == emoji:
                self.match = func
                return True
        return False

    async def paginate(self):
        """Actually paginate the entries and run the interactive loop if necessary."""
        await self.show_page(1, first=True)

        while self.paginating:
            react = await self.bot.wait_for_reaction(message=self.message, check=self.react_check, timeout=120.0)
            if react is None:
                self.paginating = False
                try:
                    await self.bot.clear_reactions(self.message)
                except:
                    pass
                finally:
                    break

            try:
                await self.bot.remove_reaction(self.message, react.reaction.emoji, react.user)
            except:
                pass # can't remove it so don't bother doing so

            await self.match()



import discord
from discord.embeds import EmptyEmbed
import itertools
import inspect

from discord.ext.commands.core import GroupMixin, Command
from discord.ext.commands import CommandError
from discord.ext.commands.formatter import HelpFormatter, Paginator

def build_embed(**params):
    title = params.get("title", EmptyEmbed)
    description = params.get("description", EmptyEmbed)
    color = params.get("color", EmptyEmbed)
    url = params.get("url", EmptyEmbed)
    author = params.get("author", "")
    author_url = params.get("author_url", EmptyEmbed)
    author_img = params.get("author_img", EmptyEmbed)
    footer = params.get("footer", "")
    footer_img = params.get("footer_img", EmptyEmbed)
    timestamp = params.get("timestamp", EmptyEmbed)
    image = params.get("image", "")
    thumbnail = params.get("thumbnail", "")
    sections = params.get("sections", params.get("fields", []))
    e = discord.Embed()
    e.title = title
    e.description = description
    e.colour = color
    e.url = url
    if author:
        e.set_author(name=author, url=author_url, icon_url=author_img)
    if footer:
        e.set_footer(text=footer, icon_url=footer_img)
    e.timestamp = timestamp
    e.set_image(url=image)
    e.set_thumbnail(url=thumbnail)
    if sections:
        populate(e, sections)
    return e


def populate(embed : discord.Embed, sections : list):
    for section in sections:
        name = section.get("name", "")
        value = section.get("value", "")
        inline = section.get("inline", True)
        if not name or not value:
            continue
        embed.add_field(name=name, value=value, inline=inline)


class EmbedHelpFormatter(HelpFormatter):
    """The default base implementation that handles formatting of the help
    command.
    To override the behaviour of the formatter, :meth:`format`
    should be overridden. A number of utility functions are provided for use
    inside that method.
    Parameters
    -----------
    bot : DiscordBot
        Bot which will be using this formatter.
    show_hidden : bool
        Dictates if hidden commands should be shown in the output.
        Defaults to ``False``.
    show_check_failure : bool
        Dictates if commands that have their :attr:`Command.checks` failed
        shown. Defaults to ``False``.
    width : int
        The maximum number of characters that fit in a line.
        Defaults to 80.
    """
    def __init__(self, bot, show_hidden=False, show_check_failure=False, width=80):
        super().__init__(show_hidden=show_hidden, show_check_failure=show_check_failure, width=width)
        self.bot = bot

    def get_ending_note(self):
        command_name = self.context.invoked_with
        return "Type {0}{1} command for more info on a command.\n" \
               "You can also type {0}{1} category for more info on a category.".format(self.clean_prefix, command_name)

    def format(self):
        """Handles the actual behaviour involved with formatting.
        To change the behaviour, this method should be overridden.
        Returns
        --------
        list
            A paginated output of the help command.
        """

        values = {}
        title = "Description"
        description = self.command.description + "\n\n" + self.get_ending_note() if not self.is_cog() else inspect.getdoc(self.command)
        sections = []

        if isinstance(self.command, Command):
                description = self.command.short_doc
                sections = [{"name": "Usage", "value": self.get_command_signature()},
                            {"name": "More Info", "value": self.command.help.replace(self.command.short_doc, "").format(prefix=self.clean_prefix),
                             "inline": False}]

        def category(tup):
            cog = tup[1].cog_name
            return cog + ':' if cog is not None else '\u200bNo Category:'

        if self.is_bot():
            title = self.bot.user.display_name + " Help"
            data = sorted(self.filter_command_list(), key=category)
            for category, commands in itertools.groupby(data, key=category):
                section = {}
                commands = list(commands)
                if len(commands) > 0:
                    section['name'] = category
                section['value'] = self.add_commands(commands)
                section['inline'] = False
                sections.append(section)
        elif not sections or self.has_subcommands():
            section = {"name": "Commands:", "inline": False, "value": self.add_commands(self.filter_command_list())}
            sections.append(section)

        values['title'] = title
        values['description'] = description
        values['sections'] = sections
        return values

    def add_commands(self, commands):
        value = ""
        for name, command in commands:
            if name in command.aliases:
                # skip aliases
                continue

            value += name + " - " + command.short_doc + "\n"
        return value



from discord.ext import commands

from ..bot_utils import config, checks
from ..bot_utils.paginator import Pages


class Reactions:
    """React to different phrases with a response and emoji reactions."""

    def __init__(self, bot):
        self.bot = bot
        self.config = config.Config('reactions.json', loop=bot.loop, directory="data")

    async def on_message(self, message):
        if message.author.bot:
            return

        if message.channel.is_private:
            return

        if message.channel.id in self.bot.get_cog("BotAdmin").config.get('ignored', []):
            return

        if not message.content.startswith(self.bot.command_prefix):
            await self._check_for_reactions(message)

    async def _check_for_reactions(self, message):
        reactions = self.config.get(message.server.id, {})
        activated = [s for s in message.content.lower().split() if s in reactions.keys()]
        activated = []
        msg = message.content.lower()
        for r in reactions.keys():
            location = msg.find(r.lower())
            if location >= 0:
                if location + len(r) >= len(msg):
                    activated.append(r)
                elif msg[location + len(r)] in [' ', '.', '!', '?', ',', "'"]:
                    activated.append(r)
        if activated:
            for reaction in activated:
                response = reactions.get(reaction, {}).get("response", "")
                reacts = reactions.get(reaction, {}).get("reaction", [])
                if response:
                    await self.bot.send_message(message.channel, response)

                if reacts:
                    for r in reacts:
                        await self.bot.add_reaction(message, r)

    @commands.command(pass_context=True, no_pm=True, aliases=["acr"])
    @checks.mod_or_permissions(manage_messages=True)
    async def addreaction(self, ctx, *, reactor : str):
        """Interactively adds a custom reaction"""

        data = self.config.get(ctx.message.server.id, {})
        keyword = data.get(reactor, {})

        if keyword:
            await self.bot.responses.failure(message="Reaction '{}' already exists.".format(reactor))
            return

        await self.bot.say("Okay, I'll react to '{}'. What do you want me to say? (Type $none for no response)".format(reactor))
        response = await self.bot.wait_for_message(author=ctx.message.author)

        reactions = []
        def check(reaction, user):
            if str(reaction.emoji) != "\U000023f9":
                reactions.append(reaction.emoji)
                return False
            else:
                return user == ctx.message.author

        msg = await self.bot.say("Awesome! Now react to this message any reactions I should have to '{}'. (React \U000023f9 to stop)".format(reactor))
        await self.bot.wait_for_reaction(message=msg, check=check)

        for i, reaction in enumerate(reactions):
            reaction = reaction if isinstance(reaction, str) else reaction.name + ":" + str(reaction.id)
            await self.bot.add_reaction(ctx.message, reaction)
            reactions[i] = reaction

        if response:
            keyword["response"] = response.content if response.content.lower() != "$none" else ""
        keyword["reaction"] = reactions
        data[reactor] = keyword
        await self.config.put(ctx.message.server.id, data)

        await self.bot.responses.success(message="Reaction '{}' has been added.".format(reactor))

    @commands.command(pass_context=True, no_pm=True, aliases=["lcr"])
    @checks.mod_or_permissions(manage_messages=True)
    async def listreactions(self, ctx):
        """Lists all the reactions for the server"""
        data = self.config.get(ctx.message.server.id, {})
        if not data:
            await self.bot.responses.failure(message="There are no reactions on this server.")
            return
        try:
            pager = Pages(self.bot, message=ctx.message, entries=list(data.keys()))
            pager.embed.colour = 0x738bd7  # blurple
            pager.embed.set_author(name=ctx.message.server.name + " Reactions", icon_url=ctx.message.server.icon_url)
            await pager.paginate()

        except Exception as e:
            await self.bot.say(e)

    @commands.command(pass_context=True, no_pm=True, aliases=["vcr"])
    @checks.mod_or_permissions(manage_messages=True)
    async def viewreaction(self, ctx, *, reactor : str):
        """Views a specific reaction"""
        data = self.config.get(ctx.message.server.id, {})
        keyword = data.get(reactor, {})

        if not keyword:
            await self.bot.responses.failure(message="Reaction '{}' was not found.".format(reactor))
            return

        response = data.get(reactor, {}).get("response", "")
        reacts = data.get(reactor, {}).get("reaction", [])

        for i, r in enumerate(reacts):
            if ":" in r:
                reacts[i] = "<:" + r + ">"

        reacts = " ".join(reacts) if reacts else "-"
        response = response if response else "-"

        string = "Here's what I say to '{reactor}': {response}\n"\
                 "I'll react to this message how I react to '{reactor}'.".format(reactor=reactor,response=response)

        await self.bot.responses.full(sections=[{"name": "Response", "value": response},
                                                {"name": "Reactions", "value": reacts, "inline": False}])



    @commands.command(pass_context=True, no_pm=True, aliases=["dcr"])
    @checks.mod_or_permissions(manage_messages=True)
    async def deletereaction(self, ctx, *, reactor : str):
        """Removes a reaction"""
        data = self.config.get(ctx.message.server.id, {})
        keyword = data.get(reactor, {})
        if keyword:
            data.pop(reactor)
            await self.config.put(ctx.message.server.id, data)
            await self.bot.responses.success(message="Reaction '{}' has been deleted.".format(reactor))
        else:
            await self.bot.responses.failure(message="Reaction '{}' was not found.".format(reactor))

    @commands.command(pass_context=True, no_pm=True, aliases=["dcrall"])
    @checks.mod_or_permissions(manage_messages=True)
    async def deleteallreactions(self, ctx):
        """Removes a reaction"""
        data = self.config.get(ctx.message.server.id, {})
        if data:
            await self.config.put(ctx.message.server.id, {})
            await self.bot.responses.success(message="All reactions have been deleted.")
        else:
            await self.bot.responses.failure(message="This server has no reactions.")

def setup(bot):
    bot.add_cog(Reactions(bot))



import discord.utils
from discord.ext import commands

from discordbot.bot_utils import config

settings = config.Config('settings.json', directory="")
meta = settings.get('meta', {})
owners = meta.get('owner', "249746236008169473")


def is_owner_check(message):
    return message.author.id == owners

def is_owner():
    return commands.check(lambda ctx: is_owner_check(ctx.message))

# The permission system of the bot is based on a "just works" basis
# You have permissions and the bot has permissions. If you meet the permissions
# required to execute the command (and the bot does as well) then it goes through
# and you can execute the command.
# If these checks fail, then there are two fallbacks.
# A role with the name of Bot Mod and a role with the name of Bot Admin.
# Having these roles provides you access to certain commands without actually having
# the permissions required for them.
# Of course, the owner will always be able to execute commands.

def check_permissions(ctx, perms):
    msg = ctx.message
    if is_owner_check(msg):
        return True

    ch = msg.channel
    author = msg.author
    resolved = ch.permissions_for(author)
    return all(getattr(resolved, name, None) == value for name, value in perms.items())

def role_or_permissions(ctx, check, **perms):
    if check_permissions(ctx, perms):
        return True

    ch = ctx.message.channel
    author = ctx.message.author
    if ch.is_private:
        return False # can't have roles in PMs

    role = discord.utils.find(check, author.roles)
    return role is not None

def mod_or_permissions(**perms):
    def predicate(ctx):
        return role_or_permissions(ctx, lambda r: r.name in ('Bot Mod', 'Bot Admin'), **perms)

    return commands.check(predicate)

def admin_or_permissions(**perms):
    def predicate(ctx):
        return role_or_permissions(ctx, lambda r: r.name == 'Bot Admin', **perms)

    return commands.check(predicate)

def is_in_servers(*server_ids):
    def predicate(ctx):
        server = ctx.message.server
        if server is None:
            return False
        return server.id in server_ids
    return commands.check(predicate)



from enum import Enum

class Colors(Enum):
    success = 0x71cd40
    failure = 0xe74a3c
    blurple = 0x738bd7

    @staticmethod
    def generate_color(red, green, blue):
        return int("0x{:02X}{:02X}{:02X}".format(red, green, blue), base=16)

    @staticmethod
    def get_default(bot):
        color = bot.config.get("meta", {}).get("default_color", "#738bd7").replace("#", "")
        return int(color, base=16)


import discordbot
from . import embeds
from .colors import Colors


class Messages:

    def __init__(self, bot : discordbot):
        self.bot = bot
        self.default_color = self.bot.config.get("meta", {}).get("default_color", "#738bd7").replace("#", "")
        self.default_color = int(self.default_color, base=16)

    async def say(self, **params):
        embed = params.get("embed", None)
        destination = params.get("destination", None)
        if not embed:
            return

        if destination:
            return await self.bot.send_message(destination, embed=embed)
        else:
            return await self.bot.say(embed=embed)

    async def direct_message(self, **params):
        member = params.get("member", "")
        if not member:
            return
        e = embeds.build_embed(description=params.pop("message", ""), author=member.display_name, author_img=member.avatar_url,
                               color=params.pop("color", Colors.get_default(self.bot)), **params)
        return await self.say(**params, embed=e)

    async def full(self, **params):
        params["color"] = params.get("color", Colors.get_default(self.bot))
        e = embeds.build_embed(**params)
        return await self.say(**params, embed=e)

    async def basic(self, **params):
        color = params.pop("color", Colors.get_default(self.bot))
        if isinstance(color, Colors): color = color.value
        e = embeds.build_embed(title=params.pop("title", ""), description=params.pop("message", params.pop("description", "")),
                               color=params.pop("color", Colors.get_default(self.bot)), **params)
        return await self.say(**params, embed=e)

    async def success(self, **params):
        title = params.pop("title", "")
        message = params.pop("message", "Success")
        return await self.basic(title=title, message=message, color=Colors.success, **params)

    async def failure(self, **params):
        title = params.pop("title", "")
        message = params.pop("message", "Success")
        return await self.basic(title=title, message=message, color=Colors.failure, **params)

    async def toggle(self, **params):
        title = params.pop("title", "")
        success = params.pop("success", True)
        if success:
            fmt = "enabled"
        else:
            fmt = "disabled"
        message = params.pop("message", "{status}").format(status=fmt)

        if success:
            return await self.success(title=title, message=message, **params)
        else:
            return await self.failure(title=title, message=message, **params)


import logging
import os
import sys

import discord
from discord.ext import commands
from discord.ext.commands.bot import _mention_pattern, _mentions_transforms

from . import colors, embeds
from .bot_utils import config
from .messages import Messages
import traceback, datetime


async def _default_help_command(ctx, *commands : str):
    """Shows this message."""
    bot = ctx.bot
    destination = ctx.message.author if bot.pm_help else ctx.message.channel

    def repl(obj):
        return _mentions_transforms.get(obj.group(0), '')

    # help by itself just lists our own commands.
    if len(commands) == 0:
        pages = bot.formatter.format_help_for(ctx, bot)
    elif len(commands) == 1:
        # try to see if it is a cog name
        name = _mention_pattern.sub(repl, commands[0])
        command = None
        if name in [x.lower() for x in bot.cogs]:
            command = bot.cogs[[x for x in bot.cogs if x.lower() == name][0]]
        else:
            command = bot.commands.get(name)
            if command is None:
                await bot.responses.failure(destination=destination, message=bot.command_not_found.format(name))
                return

        pages = bot.formatter.format_help_for(ctx, command)
    else:
        name = _mention_pattern.sub(repl, commands[0])
        command = bot.commands.get(name)
        if command is None:
            await bot.responses.failure(destination=destination, message=bot.command_not_found.format(name))
            return

        for key in commands[1:]:
            try:
                key = _mention_pattern.sub(repl, key)
                command = command.commands.get(key)
                if command is None:
                    await bot.responses.failure(destination=destination, message=bot.command_not_found.format(name))
                    return
            except AttributeError:
                await bot.responses.failure(destination=destination, message=bot.command_has_no_subcommands.format(command, key))
                return

        pages = bot.formatter.format_help_for(ctx, command)

    if bot.pm_help is None:
        characters = sum(map(lambda l: len(l), pages.values()))
        if characters > 1000:
            destination = ctx.message.author

    await bot.responses.full(destination=destination, **pages)

class DiscordBot(commands.Bot):

    def __init__(self, command_prefix=None, formatter=None, description=None, pm_help=None, **options):
        self.config = config.Config('settings.json', directory="")

        if command_prefix is None:
            command_prefix = self.config.get("meta", {}).get("prefix", "")
        if not command_prefix:
            print("Prefix was not supplied")
            sys.exit(1)

        if description is None:
            description = self.config.get("meta", {}).get("description", "")

        if pm_help is None:
            pm_help = self.config.get("meta", {}).get("pm_help", None)

        help_attrs = options.pop("help_attrs", {})
        help_attrs['hidden'] = True

        super().__init__(command_prefix, formatter=formatter if formatter else embeds.EmbedHelpFormatter(self),
                         description=description, pm_help=pm_help, command_not_found='No command called "{}" found.',
                         command_has_no_subcommands='Command "{0.name}" has no subcommands.', help_attrs=help_attrs, **options)
        os.makedirs("logs", exist_ok=True)
        discord_logger = logging.getLogger('discord')
        discord_logger.setLevel(logging.CRITICAL)
        discord_logger.addHandler(logging.FileHandler(filename='logs/errors.log', encoding='utf-8'))

        stats_log = logging.getLogger('stats')
        stats_log.setLevel(logging.INFO)
        stats_log.addHandler(logging.FileHandler(filename='logs/stats.log', encoding='utf-8'))

        info_log = logging.getLogger('info')
        info_log.setLevel(logging.INFO)
        info_log.addHandler(logging.FileHandler(filename='logs/info.log', encoding='utf-8'))

        self.logs = {'discord': discord_logger, 'stats': stats_log, 'info': info_log}

        credentials = self.config.get('credentials', {})

        client_id = credentials.get('client_id', "")
        token = credentials.get('token', "")

        if not client_id or not token:
            print("Could not find both client_id and token in settings.json")
            sys.exit(1)

        self.ownerInviteLink = self.config.get("meta", {}).get("invite_link", "")
        self.client_id = client_id
        self.token = token
        self.colors = colors.Colors
        self.responses = Messages(self)

        self.remove_command("help")
        self.command(**self.help_attrs)(_default_help_command)

    async def set_prefix(self, prefix):
        self.command_prefix = prefix
        await self.change_presence(game=discord.Game(name='{}help for help'.format(prefix)))

    def load_cogs(self, cogs = None):
        if cogs is None:
            cogs = self.config.get("cogs", [])
        cogs.extend(['discordbot.cogs.meta', 'discordbot.cogs.botadmin'])
        for extension in cogs:
            try:
                self.load_extension(extension)
            except Exception as e:
                print('Failed to load extension {}\n{}: {}'.format(extension, type(e).__name__, e))

    def run(self):
        super().run(self.token)

    async def on_command_error(self, error, ctx):
        if isinstance(error, commands.NoPrivateMessage):
            await self.send_message(ctx.message.author, 'This command cannot be used in private messages.')
        elif isinstance(error, commands.DisabledCommand):
            await self.send_message(ctx.message.author, 'Sorry. This command is disabled and cannot be used.')
        elif isinstance(error, commands.CommandInvokeError):
            print('In {0.command.qualified_name}:'.format(ctx), file=sys.stderr)
            traceback.print_tb(error.original.__traceback__)
            print('{0.__class__.__name__}: {0}'.format(error.original), file=sys.stderr)

    async def on_ready(self):
        print('Logged in as:')
        print('Username: ' + self.user.name)
        print('ID: ' + self.user.id)
        print('------')
        await self.change_presence(game=discord.Game(name='{}help for help'.format(self.command_prefix)))
        if not hasattr(self, 'uptime'):
            self.uptime = datetime.datetime.utcnow()

    async def on_message(self, message):
        if message.author.bot:
            return

        await self.process_commands(message)

    async def on_resumed(self):
        print('resumed...')

    async def logout(self):
        await super(DiscordBot, self).logout()
        for log in self.logs.values():
            for h in log.handlers:
                h.close()
                log.removeHandler(h)


async def entry_to_code(bot, entries):
    width = max(map(lambda t: len(t[0]), entries))
    output = ['```']
    fmt = '{0:<{width}}: {1}'
    for name, entry in entries:
        output.append(fmt.format(name, entry, width=width))
    output.append('```')
    await bot.say('\n'.join(output))

import datetime

async def indented_entry_to_code(bot, entries):
    width = max(map(lambda t: len(t[0]), entries))
    output = ['```']
    fmt = '\u200b{0:>{width}}: {1}'
    for name, entry in entries:
        output.append(fmt.format(name, entry, width=width))
    output.append('```')
    await bot.say('\n'.join(output))

async def too_many_matches(bot, msg, matches, entry):
    check = lambda m: m.content.isdigit()
    await bot.say('There are too many matches... Which one did you mean? **Only say the number**.')
    await bot.say('\n'.join(map(entry, enumerate(matches, 1))))

    # only give them 3 tries.
    for i in range(3):
        message = await bot.wait_for_message(author=msg.author, channel=msg.channel, check=check)
        index = int(message.content)
        try:
            return matches[index - 1]
        except:
            await bot.say('Please give me a valid number. {} tries remaining...'.format(2 - i))

    raise ValueError('Too many tries. Goodbye.')

class Plural:
    def __init__(self, **attr):
        iterator = attr.items()
        self.name, self.value = next(iter(iterator))

    def __str__(self):
        v = self.value
        if v > 1:
            return '%s %ss' % (v, self.name)
        return '%s %s' % (v, self.name)

def human_timedelta(dt):
    now = datetime.datetime.utcnow()
    delta = now - dt
    hours, remainder = divmod(int(delta.total_seconds()), 3600)
    minutes, seconds = divmod(remainder, 60)
    days, hours = divmod(hours, 24)
    years, days = divmod(days, 365)

    if years:
        if days:
            return '%s and %s ago' % (Plural(year=years), Plural(day=days))
        return '%s ago' % Plural(year=years)

    if days:
        if hours:
            return '%s and %s ago' % (Plural(day=days), Plural(hour=hours))
        return '%s ago' % Plural(day=days)

    if hours:
        if minutes:
            return '%s and %s ago' % (Plural(hour=hours), Plural(minute=minutes))
        return '%s ago' % Plural(hour=hours)

    if minutes:
        if seconds:
            return '%s and %s ago' % (Plural(minute=minutes), Plural(second=seconds))
        return '%s ago' % Plural(minute=minutes)
    return '%s ago' % Plural(second=seconds)

#
# Created by Zack Rauen
#
def title_case(string : str):
    return ' '.join([s.strip().capitalize() for s in string.split(' ')]).strip()

def sentence_case(string : str):
    return '. '.join([s.strip().capitalize() for s in string.split('.')]).strip()


# -*- coding: utf-8 -*-

"""
Discord API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the Discord.py library.
:copyright: (c) 2015-2016 Zack Rauen
:license: MIT, see LICENSE for more details.
"""

__title__ = 'discordbot'
__author__ = 'Zack Rauen'
__license__ = 'MIT'
__copyright__ = 'Copyright 2017 Zack Rauen'
__version__ = '0.2.3.a2'


from .discordbot import DiscordBot
from .colors import Colors
from .utilities import *
from .bot_utils.config import Config
from .bot_utils.paginator import Pages
from .bot_utils.formats import human_timedelta
from .bot_utils import checks, config, paginator, formats
from discord.ext import commands
from discord.ext.commands import *
from discord import *


from collections import namedtuple
VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')

version_info = VersionInfo(major=0, minor=1, micro=9, releaselevel='alpha', serial=0)


from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='discordbot.py',
    version='0.2.3.a3',
    description='A wrapper for discord.py with advanced functionality',
    long_description=long_description,
    url='https://github.com/rauenzi/discordbot.py',
    author='Zack Rauen',
    author_email='rauenzi@outlook.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
    keywords='discord framework development',
    packages=find_packages(),
    install_requires=['discord.py', 'psutil'],
    python_requires=">=3.5",
    extras_require={
        'voice': ['discord.py[voice]']
    }
)



from difflib import SequenceMatcher
import aiohttp
import re

def wordcount(string : str):
    return len(string.split())

def escape_markdown(string : str):
    return re.sub("([*~_`])", r"\\\1", string)

def similar(a, b):
    return SequenceMatcher(None, a, b).ratio()


def constant(f):
    def fset(self, value):
        raise TypeError
    def fget(self):
        return f()
    return property(fget, fset)

async def fetch(url, loop):
    async with aiohttp.ClientSession(loop=loop) as session:
        with aiohttp.Timeout(10, loop=session.loop):
            async with session.get(url) as response:
                return await response.text()